//@ts-check
"use strict";

/**
@typedef AmmoDesc
@type {{
  id: AmmoName,
  lifespan: number;
  cooldown: number;
  speed: number;
  damage: number;
  sprite?: string;
  spawns?: () => NPCDesc;
}}
*/

/**
@typedef AmmoName
@type {keyof typeof ammoDescs}
 */

/**
 * @type {{
 *   pea: AmmoDesc,
 *   goo: AmmoDesc,
 *   onion: AmmoDesc,
 *   leek: AmmoDesc,
 *   salad: AmmoDesc,
 * }}
 */
const ammoDescs = {
  pea: {
    id: "pea",
    lifespan: 1,
    cooldown: 0.1,
    speed: 12,
    damage: 1,
  },
  goo: {
    id: "goo",
    lifespan: 2,
    cooldown: 0.4,
    speed: 6,
    damage: 3,
  },
  onion: {
    id: "onion",
    lifespan: 0.5,
    cooldown: 1,
    speed: 8,
    damage: 1,
    sprite: "minion",
    spawns: () => npcDescs.onion,
  },
  leek: {
    id: "leek",
    lifespan: 0.5,
    cooldown: 1,
    speed: 8,
    damage: 1,
    sprite: "minion",
    spawns: () => npcDescs.leek,
  },
  salad: {
    id: "salad",
    lifespan: 0.5,
    cooldown: 1,
    speed: 8,
    damage: 1,
    sprite: "minion",
    spawns: () => npcDescs.salad,
  },
};


/**
@typedef Bullet
@type {{
  pos: PIXI.Point;
  velocity: PIXI.Point;
  lifespan: number;
  ammoDesc: AmmoDesc;
  sprite: PIXI.AnimatedSprite;
}}
*/

/**
 * @param {number} dt
 */
function updateBullets(dt) {
  const { bullets } = inGame.activeWorld;

  let i = 0;
  while (i < bullets.length) {
    const bullet = bullets[i];
    bullet.lifespan -= dt;
    if (bullet.lifespan <= 0) {
      onBulletExplode(bullet);
      continue;
    }

    const nextBulletPos = new PIXI.Point(bullet.pos.x + dt * bullet.velocity.x, bullet.pos.y + dt * bullet.velocity.y);

    // Check collision with enemies
    /** @type {Baddie} */
    let closestHitEnemy = null;
    let closestHitEnemyDistance = Infinity;
    let closestHitEnemeyIntersection;

    for (let enemy of inGame.activeWorld.baddies) {
      const enemyHitbox = new PIXI.Rectangle(enemy.pos.x - enemy.hitbox.width / 2, enemy.pos.y - enemy.hitbox.height, enemy.hitbox.width, enemy.hitbox.height);

      if (enemyHitbox.contains(bullet.pos.x, bullet.pos.y)) {
        closestHitEnemyDistance = 0;
        closestHitEnemy = enemy;
        closestHitEnemeyIntersection = bullet.pos;
        break;
      }

      const intersection = findSegmentRectIntersection(bullet.pos, nextBulletPos, enemyHitbox);

      if (intersection != null) {
        const dist = distxy(bullet.pos, intersection);
        if (dist < closestHitEnemyDistance) {
          closestHitEnemyDistance = dist;
          closestHitEnemy = enemy;
          closestHitEnemeyIntersection = intersection;
        }
      }
    }

    const angle = Math.atan2(-bullet.velocity.y, -bullet.velocity.x)

    // Check collision with map
    // TODO: Do a line algo to check every tile in between instead of just the last one? If it's too fast it might go through walls right now
    const tileCollision = getTileCollisionAt(nextBulletPos.x, nextBulletPos.y);
    if (tileCollision != null && !tileCollision.ignoreBullets) {
      const tileIntersection = findSegmentRectIntersection(bullet.pos, nextBulletPos, tileCollision.rect);

      if (tileIntersection != null) {
        const tileDist = distxy(bullet.pos, tileIntersection);

        if (closestHitEnemy == null || closestHitEnemyDistance > tileDist) {
          onBulletExplode(bullet);
          play2Dsound(`hits.wall`, tileIntersection);

          const spriteName = bullet.ammoDesc.sprite != null ? bullet.ammoDesc.sprite : bullet.ammoDesc.id;
          spawnAnimatedFx(`inGame/bullets/${spriteName}Hit`, tileIntersection, angle);

          continue;
        }
      }
    }

    if (closestHitEnemy != null) {
      onBulletExplode(bullet);
      damageBaddie(closestHitEnemy, bullet.ammoDesc.damage, angle);
      play2Dsound(`hits.baddies.${closestHitEnemy.baddieDesc.id}`, closestHitEnemeyIntersection);

      const spriteName = bullet.ammoDesc.sprite != null ? bullet.ammoDesc.sprite : bullet.ammoDesc.id;
      spawnAnimatedFx(`inGame/bullets/${spriteName}Hit`, closestHitEnemeyIntersection, angle);

      continue;
    }

    bullet.pos.copyFrom(nextBulletPos);
    bullet.sprite.position.copyFrom(bullet.pos);

    bullet.sprite.update(app.ticker.deltaTime);

    i++;
  }
}

/**
 * Spawns a new bullet instance
 * @param {PIXI.Point} pos
 * @param {PIXI.Point} velocity
 * @param {AmmoDesc} ammoDesc - Ammo being fired
 */
function spawnBullet(pos, velocity, ammoDesc) {
  const angle = Math.atan2(-velocity.y, -velocity.x);

  const spriteName = ammoDesc.sprite != null ? ammoDesc.sprite : ammoDesc.id;

  spawnAnimatedFx(`inGame/bullets/${spriteName}Blast`, pos, angle);

  const spritesheet = getSpritesheet(`inGame/bullets/${spriteName}`);
  const sprite = new PIXI.AnimatedSprite(spritesheet.animations["anim"], false);
  sprite.anchor.copyFrom(spritesheet.anchor);
  sprite.scale.set(inverseTileSize, inverseTileSize);
  sprite.angle = angle * PIXI.RAD_TO_DEG + 180; // + 180 because sprites are looking left instead of right

  sprite.play();

  inGame.activeWorld.pixiLayers["ground"].particlesContainer.addChild(sprite);

  /** @type Bullet */
  const bullet = {
    pos,
    velocity,
    lifespan: ammoDesc.lifespan,
    ammoDesc,

    sprite
  };

  inGame.activeWorld.bullets.push(bullet);
}

/**
 * @param {Bullet} bullet
 */
function onBulletExplode(bullet) {
  inGame.activeWorld.pixiLayers["ground"].particlesContainer.removeChild(bullet.sprite);
  inGame.activeWorld.bullets.splice(inGame.activeWorld.bullets.indexOf(bullet), 1);

  bullet.sprite.destroy();

  if (bullet.ammoDesc.spawns != null) {
    const tileCollision = getTileCollisionAt(bullet.pos.x, bullet.pos.y);
    if (tileCollision == null) {
      createNPC(inGame.activeWorld, bullet.pos, bullet.ammoDesc.spawns(), true /* following */)
    }
  }
}
